/********************************************************/
/*  ntapi: Native API core library                      */
/*  Copyright (C) 2013--2017  Z. Gilboa                 */
/*  Released under GPLv2 and GPLv3; see COPYING.NTAPI.  */
/********************************************************/

#include <psxtypes/psxtypes.h>
#include <ntapi/nt_file.h>
#include <ntapi/nt_string.h>
#include <ntapi/nt_atomic.h>
#include <ntapi/nt_port.h>
#include <ntapi/nt_ipc.h>
#include <ntapi/nt_msq.h>
#include <ntapi/ntapi.h>
#include "ntapi_impl.h"

static int32_t __msq_send_return(
	intptr_t *	hlock,
	int32_t		status)
{
	at_store(hlock,0);
	return status;
}


int32_t __stdcall __ntapi_msq_send(
	__in	nt_msq_info *		msq,
	__in 	void *			hevent		__optional,
	__in	nt_io_apc_routine *	apc_routine	__optional,
	__in	void * 			apc_context	__optional,
	__in	const void *		buffer,
	__in	size_t			len,
	__in	intptr_t		rank,
	__in	uint32_t		options,
	__out	nt_io_status_block *	iosb)
{
	int32_t			status;
	void *			hsection;
	void *			secaddr;
	size_t			secsize;
	nt_msq_info_msg		msg;
	intptr_t *		hlock;

	/* validate */
	if (!iosb)
		return NT_STATUS_INVALID_PARAMETER;

	/* section */
	if (msq->section_addr) {
		hsection = msq->section;
		secaddr  = msq->section_addr;
		secsize  = msq->section_size;

	} else if ((status = __ntapi->ipc_init_section_by_port(
			msq->hport,&hsection,
			&secaddr,&secsize)))
		return status;

	if (len > secsize)
		return NT_STATUS_INFO_LENGTH_MISMATCH;

	/* lock */
	hlock = &(__ntapi_internals()->hlock);

	if (at_locked_cas(hlock,0,1))
		return NT_STATUS_RESOURCE_NOT_OWNED;

	/* msq data to section */
	__ntapi->tt_generic_memcpy(
		secaddr,buffer,len);

	/* msg */
	__ntapi->tt_aligned_block_memset(
		&msg,0,sizeof(msg));

	msg.header.msg_type		= NT_LPC_NEW_MESSAGE;
	msg.header.data_size		= sizeof(msg.data);
	msg.header.msg_size		= sizeof(msg);
	msg.data.ttyinfo.opcode		= NT_TTY_MSQ_SEND;

	msg.data.msqinfo.msqkey		= msq->msqkey;
	msg.data.msqinfo.msqid		= msq->msqid;
	msg.data.msqinfo.msqspid	= msq->msqspid;
	msg.data.msqinfo.msqrank	= rank;
	msg.data.msqinfo.section_addr	= secaddr;
	msg.data.msqinfo.section_size	= len;
	msg.data.msqinfo.ntoptions	= options;

	msg.data.msqinfo.hevent		= hevent;
	msg.data.msqinfo.apc_routine	= apc_routine;
	msg.data.msqinfo.apc_context	= apc_context;
	msg.data.msqinfo.riosb		= iosb;

	if ((status = __ntapi->zw_request_wait_reply_port(msq->hport,&msg,&msg)))
		return __msq_send_return(hlock,status);
	else if (msg.data.ttyinfo.status)
		return __msq_send_return(hlock,msg.data.ttyinfo.status);

	iosb->status = msg.data.msqinfo.ntiosb.status;
	iosb->info   = msg.data.msqinfo.ntiosb.info;

	return __msq_send_return(hlock,NT_STATUS_SUCCESS);
}
